import asyncio
import functools
import logging

import bluepy.btle
import btlewrap
from btlewrap import BluetoothBackendException
from fhempy.lib.generic import FhemModule
from miflora import miflora_poller

from .. import fhem, utils


class miflora(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self.hash = None
        self.updateTask = None
        self._poller = None
        self._attr_update_interval = 1200
        self._attr_hci_device = "hci0"
        self._attr_list = {
            "update_interval": {"default": 1200, "format": "int"},
            "hci_device": {"default": "hci0"},
            "poll_type": {"default": "interval", "options": "interval,manual"},
        }
        self.set_attr_config(self._attr_list)
        self._set_list = {"update": {}}
        self.set_set_config(self._set_list)

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 4:
            return "Usage: define mi_plant PythonModule miflora <MAC>"

        self._address = args[3]
        hash["MAC"] = args[3]
        self.logger.debug(f"Define miflora: {self._address}")

        self._poller = miflora_poller.MiFloraPoller(
            self._address,
            btlewrap.BluepyBackend,
            cache_timeout=60,
            adapter=self._attr_hci_device,
        )

        if self.updateTask:
            self.updateTask.cancel()
        self.updateTask = self.create_async_task(self.update_task())

    async def update_task(self):
        while True:
            await self.do_update()
            await asyncio.sleep(self._attr_update_interval)

    async def do_update(self):
        self.logger.debug("Run update task")
        try:
            # name
            name = await utils.run_blocking(functools.partial(self._poller.name))
            await fhem.readingsSingleUpdateIfChanged(self.hash, "name", name, 1)
            # firmware_version
            firmware_version = await utils.run_blocking(
                functools.partial(self._poller.firmware_version)
            )
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "firmware", firmware_version, 1
            )
            for param in (
                "temperature",
                "light",
                "moisture",
                "conductivity",
                "battery",
            ):
                # param
                param_val = await utils.run_blocking(
                    functools.partial(self._poller.parameter_value, param)
                )
                if param == "conductivity":
                    param = "fertility"
                await fhem.readingsSingleUpdateIfChanged(self.hash, param, param_val, 1)
            await fhem.readingsSingleUpdateIfChanged(self.hash, "presence", "online", 1)
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "online", 1)
        except Exception:
            self.logger.error(f"Failed to get updates from miflora {self._address}")
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, "presence", "offline", 1
            )
            await fhem.readingsSingleUpdateIfChanged(self.hash, "state", "offline", 1)

    async def set_update(self, hash, params):
        self.create_async_task(self.do_update())

    async def set_attr_update_interval(self, hash):
        await fhem.readingsSingleUpdateIfChanged(
            self.hash, "update_interval", str(self._attr_update_interval), 1
        )
        if self.updateTask:
            self.updateTask.cancel()
        self.updateTask = self.create_async_task(self.update_task())

    async def set_attr_poll_type(self, hash):
        if self._attr_poll_type == "manual":
            if self.updateTask:
                self.updateTask.cancel()
                self.updateTask = None
        elif self._attr_poll_type == "interval":
            if self.updateTask is None:
                self.updateTask = self.create_async_task(self.update_task())
        return

    async def set_attr_hci_device(self, hash):
        self.logger.debug("attr change of hci device")
        self._poller = miflora_poller.MiFloraPoller(
            self._address,
            btlewrap.BluepyBackend,
            cache_timeout=60,
            adapter=self._attr_hci_device,
        )



import asyncio
import concurrent.futures
import functools
import json
import logging
import threading
import time
import urllib.request
from urllib.parse import parse_qs, urlparse

import aiohttp

# PyChromecast
import pychromecast

# DashCast
import pychromecast.controllers.dashcast as dashcast
import requests

# youtube_dl
import youtube_dl

# BubbleUPNP
from pychromecast.controllers.bubbleupnp import BubbleUPNPController

# Spotify
from pychromecast.controllers.spotify import SpotifyController
import spotipy
from spotipy.oauth2 import CacheFileHandler

# YouTube
from pychromecast.controllers.youtube import YouTubeController

from .. import fhem, utils
from ..generic import FhemModule

connection_update_lock = threading.Lock()


class googlecast(FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        # do only error logging for pychromecast library
        logging.getLogger("pychromecast").setLevel(logging.ERROR)
        self.cast = None
        self.loop = asyncio.get_running_loop()
        self.hash = None
        self.currPosTask = None
        self.connectionStateCache = ""
        self.browser = None
        self.spotipy = None
        attr_conf = {
            "favorite_1": {"default": ""},
            "favorite_2": {"default": ""},
            "favorite_3": {"default": ""},
            "favorite_4": {"default": ""},
            "favorite_5": {"default": ""},
            "speak_lang": {"default": "de", "options": "de,en,nl,fr,it,es"},
            "spotify_sp_dc": {
                "default": "",
                "help": "Go to chrome://settings/cookies/detail?site=spotify.com"
                + " and copy the content of sp_dc",
                "function": "set_attr_spotify_cookie",
            },
            "spotify_sp_key": {
                "default": "",
                "help": "Go to chrome://settings/cookies/detail?site=spotify.com"
                + " and copy the content of sp_key",
                "function": "set_attr_spotify_cookie",
            },
            "player": {
                "default": "DefaultMediaRenderer",
                "help": "Play with BubbleUpnp player or"
                + " Default Media Renderer cast app",
                "options": "BubbleUpnp,DefaultMediaRenderer",
            },
        }
        self.set_attr_config(attr_conf)

        self._set_conf = {
            "authcode": {"args": ["code"]},
            "stop": {},
            "pause": {},
            "rewind": {},
            "skip": {},
            "quitApp": {},
            "play": {
                "args": ["url"],
                "argsh": ["url"],
                "params": {"url": {"optional": True, "default": ""}},
            },
            "addToQueue": {
                "args": ["url"],
                "argsh": ["url"],
                "params": {"url": {"default": ""}},
            },
            "playFavorite": {"args": ["id"], "options": "1,2,3,4,5"},
            "volume": {
                "args": ["vol"],
                "params": {"vol": {"format": "int"}},
                "options": "slider,0,1,100",
            },
            "seek": {"args": ["pos"]},
            "next": {},
            "prev": {},
            "subtitles": {"args": ["onoff"], "options": "on,off"},
            "displayWebsite": {
                "args": ["url"],
                "argsh": ["url"],
                "params": {"url": {"optional": True, "default": "https://www.fhem.de"}},
            },
            "speak": {
                "args": ["text"],
                "help": "Please provide text with quotes.",
            },
            "startApp": {"args": ["appid"]},
            "startSpotify": {},
            "volUp": {},
            "volDown": {},
        }
        self.set_set_config(
            {
                "waiting": {
                    "help": "Please wait till connected and"
                    + " reload the page when connected."
                }
            }
        )

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) > 3:
            hash["CASTNAME"] = args[3]
        else:
            return 'Usage: define my_fhempy_cast PythonModule googlecast "Living Room"'

        if self.browser:
            self.browser.stop_discovery()
        if self.cast:
            self.cast.disconnect()

        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "state", "offline")
        await fhem.readingsBulkUpdateIfChanged(hash, "connection", "disconnected")
        await fhem.readingsEndUpdate(hash, 1)
        await self.set_auth_url()
        self.create_async_task(self.connect_spotipy())

        self.startDiscovery()

    # FHEM FUNCTION
    async def Undefine(self, hash):
        await super().Undefine(hash)
        try:
            self.browser.stop_discovery()
            if self.cast:
                self.cast.disconnect()
        except Exception:
            pass

    async def set_auth_url(self):
        spotipy_scope = (
            ""
            # Spotify scopes:
            #  https://developer.spotify.com/documentation/general/guides/scopes/
            # Listening history
            + "user-read-recently-played "
            + "user-top-read "
            + "user-read-playback-position "
            # Spotify connect
            + "user-read-playback-state "
            + "user-modify-playback-state "
            + "user-read-currently-playing "
            # Playback
            + "streaming "
            # Playlist
            + "playlist-read-private "
            + "playlist-read-collaborative "
            # Follow
            + "user-follow-read "
            # Library
            + "user-library-read "
            # User
            + "user-read-email "
            + "user-read-private"
        )
        # not sure if this is useful...
        x = "e92855a009"
        y = "e74eb69ba6609d3bfd7d"
        z = 90 + 6
        handler = CacheFileHandler(cache_path=f".{self.hash['NAME']}_spotify_token")
        self.spotipy_pkce = spotipy.oauth2.SpotifyPKCE(
            f"{x}{y}{str(z)}",
            "https://europe-west1-fhem-ga-connector.cloudfunctions.net/codelanding/start",
            scope=spotipy_scope,
            cache_handler=handler,
        )
        url = self.spotipy_pkce.get_authorize_url()
        url = (
            '<html><a href="'
            + url
            + '" target="_blank">Connect Spotify account '
            + "(new window/tab)</a><br></html>"
        )
        await fhem.readingsSingleUpdate(self.hash, "spotify_login", url, 1)

    async def set_authcode(self, hash, params):
        code = params["code"]
        self.create_async_task(self.handle_authcode(code))

    async def handle_authcode(self, code):
        access_token = await utils.run_blocking(
            functools.partial(
                self.spotipy_pkce.get_access_token, code=code, check_cache=False
            )
        )
        await self.connect_spotipy()

    async def connect_spotipy(self):
        if self.spotipy_pkce.get_cached_token() is None:
            await fhem.readingsSingleUpdate(
                self.hash, "spotify_user", "login required", 1
            )
            return

        if self.spotipy is None:
            self.spotipy = spotipy.Spotify(auth_manager=self.spotipy_pkce)
            user = await utils.run_blocking(
                functools.partial(self.spotipy.current_user)
            )
            if user is not None:
                self.loggedin = True
                await fhem.readingsSingleUpdate(
                    self.hash,
                    "spotify_user",
                    user["display_name"] + " (" + user["email"] + ")",
                    1,
                )

    async def set_play(self, hash, params):
        url = params["url"]
        if url != "":
            self.playUrl(url)
        else:
            self.cast.media_controller.play()

    async def set_addToQueue(self, hash, params):
        url = params["url"]
        if url != "":
            self.queueUrl(url)

    async def set_playFavorite(self, hash, params):
        fav = params["id"]
        url = await fhem.AttrVal(self.hash["NAME"], "favorite_" + str(fav), "")
        if len(url) == 0:
            return "Please set favorite before usage"
        self.playUrl(url)

    async def set_stop(self, hash, params):
        self.cast.media_controller.stop()

    async def set_pause(self, hash, params):
        self.cast.media_controller.pause()

    async def set_quitApp(self, hash, params):
        self.cast.quit_app()

    async def set_startApp(self, hash, params):
        appId = params["appid"]
        self.cast.start_app(appId)

    async def set_skip(self, hash, params):
        self.cast.media_controller.skip()

    async def set_rewind(self, hash, params):
        self.cast.media_controller.rewind()

    async def set_seek(self, hash, params):
        position = params["pos"]
        self.cast.media_controller.seek(position)

    async def set_next(self, hash, params):
        self.cast.media_controller.queue_next()

    async def set_prev(self, hash, params):
        self.cast.media_controller.queue_prev()

    async def set_volUp(self, hash, params):
        self.cast.volume_up()

    async def set_volDown(self, hash, params):
        self.cast.volume_down()

    async def set_subtitles(self, hash, params):
        onoff = params["onoff"]
        if onoff == "on":
            self.cast.media_controller.enable_subtitle(1)
        else:
            self.cast.media_controller.disable_subtitle()

    async def set_volume(self, hash, params):
        vol = params["vol"]
        self.cast.set_volume(vol / 100)

    async def set_speak(self, hash, params):
        self.create_async_task(self.do_set_speak(hash, params))

    async def do_set_speak(self, hash, params):
        txt = params["text"]
        lang = await fhem.AttrVal(self.hash["NAME"], "speak_lang", "de")
        ttsUrl = (
            "http://translate.google.com/translate_tts?tl="
            + lang
            + "&client=tw-ob&q="
            + urllib.parse.quote(txt)
        )
        self.cast.play_media(ttsUrl, "audio/mpeg")

    async def set_displayWebsite(self, hash, params):
        dashUrl = params["url"]
        self.create_async_task(self.displayWebsite(dashUrl))

    async def set_startSpotify(self, hash, params):
        self.create_async_task(self.launchSpotify())

    def playUrl(self, url):
        videoid = self.extract_video_id(url)
        if videoid is None:
            if url.find("spotify") >= 0:
                self.create_async_task(self.playSpotify(url))
            else:
                self.create_async_task(self.playDefaultMedia(url))
        else:
            if self.cast.cast_type == "audio":
                self.create_async_task(self.playYoutubeAudio(url))
            else:
                playlistid = self.extract_playlist_id(url)
                self.create_async_task(self.playYoutube(videoid, playlistid))

    def queueUrl(self, url):
        self.create_async_task(self.playDefaultMedia(url, enqueue=True))

    async def playDefaultMedia(self, uri, enqueue=False):
        try:
            session = aiohttp.ClientSession()
            res = await session.get(uri)
            mime = res.headers["Content-Type"]
            if self._attr_player == "BubbleUpnp":
                bubbleupnp = BubbleUPNPController()
                self.cast.register_handler(bubbleupnp)
                await utils.run_blocking(functools.partial(bubbleupnp.launch))
                await utils.run_blocking(
                    functools.partial(bubbleupnp.play_media, uri, mime)
                )
            else:
                await utils.run_blocking(
                    functools.partial(self.cast.play_media, uri, mime, enqueue=enqueue)
                )
        except Exception:
            self.logger.exception(f"Failed to play: {uri}")

    async def set_attr_spotify_cookie(self, hash):
        if self._attr_spotify_sp_dc != "" and self._attr_spotify_sp_key != "":
            self.create_async_task(self.update_token())
        else:
            await fhem.readingsSingleUpdate(
                self.hash, "spotify_user", "attr spotify_sp... required", 1
            )

    def start_session(self, dc=None, key=None):
        """ Starts session to get access token. """
        USER_AGENT = (
            "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_2) "
            "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/63.0.3239.132 Safari/537.36"
        )

        session = requests.Session()

        cookies = {"sp_dc": dc, "sp_key": key}
        headers = {"user-agent": USER_AGENT}

        response = session.get(
            "https://open.spotify.com/get_access_token?"
            + "reason=transport&productType=web_player",
            headers=headers,
            cookies=cookies,
        )
        response.raise_for_status()
        data = response.content.decode("utf-8")
        config = json.loads(data)

        access_token = config["accessToken"]
        expires_timestamp = config["accessTokenExpirationTimestampMs"]
        expiration_date = int(expires_timestamp) // 1000
        return access_token, expiration_date

    async def update_token(self):
        if self._attr_spotify_sp_dc != "" and self._attr_spotify_sp_key != "":
            data = await utils.run_blocking(
                functools.partial(
                    self.start_session,
                    self._attr_spotify_sp_dc,
                    self._attr_spotify_sp_key,
                )
            )
            self.spotify_access_token = data[0]
            self.spotify_expires = data[1] - int(time.time())
            self.spotify = spotipy.Spotify(auth=self.spotify_access_token)

            user = await utils.run_blocking(
                functools.partial(self.spotify.current_user)
            )
            if user is None:
                await fhem.readingsSingleUpdate(
                    self.hash, "spotify_user", "login failed", 1
                )

    async def launchSpotify(self):
        if self.spotify_access_token is None:
            await fhem.readingsSingleUpdate(
                self.hash, "spotify_user", "attr spotify sp... and login required", 1
            )
            return
        await self.update_token()
        # Launch the spotify app on the cast we want to cast to
        sp = SpotifyController(self.spotify_access_token, self.spotify_expires)
        self.cast.register_handler(sp)
        await utils.run_blocking(functools.partial(sp.launch_app))

    async def playSpotify(self, uri):
        if self.spotify_access_token is None:
            await fhem.readingsSingleUpdate(
                self.hash, "spotify_user", "attr spotify sp... required", 1
            )
            return

        try:
            await self.update_token()
            # Launch the spotify app on the cast we want to cast to
            sp = SpotifyController(self.spotify_access_token, self.spotify_expires)
            self.cast.register_handler(sp)
            await utils.run_blocking(functools.partial(sp.launch_app))

            if not sp.is_launched and not sp.credential_error:
                self.logger.error("Failed to launch spotify controller due to timeout")
                return
            if not sp.is_launched and sp.credential_error:
                self.logger.error(
                    "Failed to launch spotify controller due to credential error"
                )
                return

            # Query spotify for active devices
            if self.spotipy is None:
                return

            devices_available = await utils.run_blocking(
                functools.partial(self.spotipy.devices)
            )

            # Match active spotify devices with the spotify controller's device id
            spotify_device_id = None
            for device in devices_available["devices"]:
                if device["id"] == sp.device:
                    spotify_device_id = device["id"]
                    break

            if not spotify_device_id:
                self.logger.error(
                    'No device with id "{}" known by Spotify'.format(sp.device)
                )
                self.logger.error(
                    "Known devices: {}".format(devices_available["devices"])
                )
                return

            # Start playback
            if uri.find("track") > 0:
                await utils.run_blocking(
                    functools.partial(
                        self.spotipy.start_playback,
                        device_id=spotify_device_id,
                        uris=[uri],
                    )
                )
            else:
                await utils.run_blocking(
                    functools.partial(
                        self.spotipy.start_playback,
                        device_id=spotify_device_id,
                        context_uri=uri,
                    )
                )
        except Exception:
            self.logger.exception("Failed to play spotify track")

    async def displayWebsite(self, url):
        d = dashcast.DashCastController()
        self.cast.register_handler(d)
        d.load_url(url, force=True, reload_seconds=30)

    async def playYoutubeAudio(self, uri):
        with concurrent.futures.ThreadPoolExecutor() as pool:
            video_url = await self.loop.run_in_executor(
                pool, functools.partial(self.getYoutubeAudioUrl, uri)
            )
            self.cast.play_media(video_url, "audio/mp4")

    def getYoutubeAudioUrl(self, uri):
        ydl = youtube_dl.YoutubeDL(
            {
                "forceurl": True,
                "simulate": True,
                "quiet": "1",
                "no_warnings": "1",
                "skip_download": True,
                "format": "bestaudio/best",
                "youtube_include_dash_manifest": True,
            }
        )
        result = ydl.extract_info(uri, download=False)
        if "entries" in result:
            # Can be a playlist or a list of videos
            video = result["entries"][0]
        else:
            # Just a video
            video = result
        video_url = video["url"]
        return video_url

    async def playYoutube(self, videoid, playlistid):
        yt = YouTubeController()
        self.cast.register_handler(yt)
        await utils.run_blocking(functools.partial(yt.play_video, videoid, playlistid))

    def extract_video_id(self, url):
        # Examples:
        # - http://youtu.be/SA2iWivDJiE
        # - http://www.youtube.com/watch?v=_oPAwA_Udwc&feature=feedu
        # - http://www.youtube.com/embed/SA2iWivDJiE
        # - http://www.youtube.com/v/SA2iWivDJiE?version=3&amp;hl=en_US
        query = urlparse(url)
        if query.hostname == "youtu.be":
            return query.path[1:]
        if query.hostname in ("www.youtube.com", "youtube.com"):
            if query.path == "/watch":
                return parse_qs(query.query)["v"][0]
            if query.path[:7] == "/embed/":
                return query.path.split("/")[2]
            if query.path[:3] == "/v/":
                return query.path.split("/")[2]
        # fail?
        return None

    def extract_playlist_id(self, url):
        query = urlparse(url)
        if "list" in parse_qs(query.query):
            return parse_qs(query.query)["list"]
        return None

    def startDiscovery(self):
        def castFound(chromecast):
            if chromecast.name == self.hash["CASTNAME"] and self.cast is None:
                self.logger.info("Discovered cast: " + chromecast.name)
                self.cast = chromecast
                # add status listener
                self.cast.register_connection_listener(self)
                self.cast.register_status_listener(self)
                # add media controller listener
                self.cast.media_controller.register_status_listener(self)
                self.logger.debug("wait for chromecast")
                # timeout 0.001 just waits for status to be ready
                # but we just need the thread to start by calling wait()
                self.cast.wait(0.001)
                self.logger.debug("wait finished")

        self.logger.debug("Start discovery")
        self.browser = pychromecast.get_chromecasts(
            blocking=False, tries=None, retry_wait=5, timeout=5, callback=castFound
        )

    # THREADING: this function is called by run_once pychromecast thread
    def new_connection_status(self, status):
        # connection update might come from different threads
        connection_update_lock.acquire()
        self.logger.debug("new_connection_status: " + status.status)

        # prevent to many disconnect events
        if status.status != self.connectionStateCache:
            self.connectionStateCache = status.status
            # run reading updates in main thread
            res = asyncio.run_coroutine_threadsafe(
                self.updateConnectionReadings(self.hash, status), self.loop
            )
            res.result()

        if status.status == "CONNECTED":
            self.set_set_config(self._set_conf)
            # run reading updates in main thread
            res = asyncio.run_coroutine_threadsafe(
                self.updateDeviceReadings(self.hash), self.loop
            )
            res.result()
        connection_update_lock.release()

    # THREADING: this function is called by run_once pychromecast thread
    def new_cast_status(self, status):
        self.logger.debug("new_cast_status")
        # run reading updates in main thread
        res = asyncio.run_coroutine_threadsafe(
            self.updateStatusReadings(self.hash, status), self.loop
        )
        res.result()

    # THREADING: this function is called by run_once pychromecast thread
    def new_media_status(self, status):
        asyncio.run_coroutine_threadsafe(
            self.stop_updateCurrentPosition(), self.loop
        ).result()
        if (
            status.player_state == "PLAYING"
            and self.currPosTask is None
            and status.duration
            and status.duration > 0
        ):
            asyncio.run_coroutine_threadsafe(
                self.run_updateCurrentPosition(), self.loop
            ).result()
        self.logger.debug("new_media_status")
        # run reading updates in main thread
        res = asyncio.run_coroutine_threadsafe(
            self.updateMediaStatusReadings(self.hash, status), self.loop
        )
        res.result()

    async def run_updateCurrentPosition(self):
        self.currPosTask = self.create_async_task(self.updateCurrentPosition())

    async def stop_updateCurrentPosition(self):
        if self.currPosTask:
            self.cancel_async_task(self.currPosTask)
            self.currPosTask = None

    async def updateCurrentPosition(self):
        try:
            while True:
                await asyncio.sleep(30)
                self.cast.media_controller.update_status()
        except concurrent.futures.CancelledError:
            # task was cancelled, nothing playing
            pass
        except Exception:
            self.logger.error("Update media status failed", exc_info=True)

    async def updateConnectionReadings(self, hash, status):
        self.logger.debug("updateConnectionReading")
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "connection", status.status.lower()
        )
        if status.status == "CONNECTED":
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "online")
        else:
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "offline")
        await fhem.readingsEndUpdate(hash, 1)

    async def updateStatusReadings(self, hash, status):
        self.logger.debug("updateStatusReadings")
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "volume", round(status.volume_level * 100)
        )
        await fhem.readingsBulkUpdateIfChanged(
            hash, "is_active_input", status.is_active_input
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "is_stand_by", status.is_stand_by)
        await fhem.readingsBulkUpdateIfChanged(hash, "mute", status.volume_muted)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "display_name", status.display_name
        )
        # await fhem.readingsBulkUpdateIfChanged(hash, "namespaces", status.namespaces)
        await fhem.readingsBulkUpdateIfChanged(hash, "session_id", status.session_id)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "transport_id", status.transport_id
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "status_text", status.status_text)
        await fhem.readingsBulkUpdateIfChanged(hash, "icon_url", status.icon_url)
        await fhem.readingsBulkUpdateIfChanged(hash, "app_id", status.app_id)
        await fhem.readingsEndUpdate(hash, 1)

    async def updateMediaStatusReadings(self, hash, status):
        self.logger.debug("updateMediaStatusReadings")
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaPlayerState", status.player_state
        )
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaContentId", status.content_id
        )
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaContentType", status.content_type
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaDuration", status.duration)
        if status.current_time:
            await fhem.readingsBulkUpdateIfChanged(
                hash, "mediaCurrentPosition", round(status.current_time)
            )
        else:
            await fhem.readingsBulkUpdateIfChanged(hash, "mediaCurrentPosition", "")
        if status.duration and status.current_time:
            await fhem.readingsBulkUpdateIfChanged(
                hash,
                "mediaCurrentPosPercent",
                round(status.current_time / status.duration * 100),
            )
        else:
            await fhem.readingsBulkUpdateIfChanged(hash, "mediaCurrentPosPercent", "")
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaStreamType", status.stream_type
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaTitle", status.title)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaSeriesTitle", status.series_title
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaSeason", status.season)
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaEpisode", status.episode)
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaArtist", status.artist)
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaAlbum", status.album_name)
        await fhem.readingsBulkUpdateIfChanged(
            hash, "mediaAlbumArtist", status.album_artist
        )
        await fhem.readingsBulkUpdateIfChanged(hash, "mediaTrack", status.track)
        if len(status.images) > 0:
            await fhem.readingsBulkUpdateIfChanged(
                hash, "mediaImageUrl", status.images[0].url
            )
            await fhem.readingsBulkUpdateIfChanged(
                hash, "mediaImageHeight", status.images[0].height
            )
            await fhem.readingsBulkUpdateIfChanged(
                hash, "mediaImageWidth", status.images[0].width
            )
        else:
            await fhem.readingsBulkUpdateIfChanged(hash, "mediaImageUrl", "")
            await fhem.readingsBulkUpdateIfChanged(hash, "mediaImageHeight", "")
            await fhem.readingsBulkUpdateIfChanged(hash, "mediaImageWidth", "")

        if status.player_state == "PLAYING":
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "playing")
        elif status.player_state == "BUFFERING":
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "buffering")
        elif status.player_state == "PAUSED":
            await fhem.readingsBulkUpdateIfChanged(hash, "state", "paused")
        else:
            if (
                await fhem.ReadingsVal(hash["NAME"], "connection", "failed")
                == "connected"
            ):
                await fhem.readingsBulkUpdateIfChanged(hash, "state", "online")
            else:
                await fhem.readingsBulkUpdateIfChanged(hash, "state", "offline")
        await fhem.readingsEndUpdate(hash, 1)

    async def updateDeviceReadings(self, hash):
        self.logger.debug("updateDeviceReadings")
        await fhem.readingsBeginUpdate(hash)
        await fhem.readingsBulkUpdateIfChanged(hash, "name", self.cast.name)
        await fhem.readingsBulkUpdateIfChanged(hash, "model_name", self.cast.model_name)
        await fhem.readingsBulkUpdateIfChanged(hash, "uuid", self.cast.uuid)
        await fhem.readingsBulkUpdateIfChanged(hash, "cast_type", self.cast.cast_type)
        # await fhem.readingsBulkUpdateIfChanged(hash, "uri", self.cast.uri)
        await fhem.readingsBulkUpdateIfChanged(hash, "ignore_cec", self.cast.ignore_cec)
        await fhem.readingsEndUpdate(hash, 1)

        if await fhem.AttrVal(hash["NAME"], "icon", "") == "":
            if await fhem.ReadingsVal(hash["NAME"], "cast_type", "cast") == "cast":
                await fhem.CommandAttr(
                    self.hash, self.hash["NAME"] + " icon scene_scene"
                )
            else:
                await fhem.CommandAttr(
                    self.hash, self.hash["NAME"] + " icon gassistant"
                )
            await fhem.CommandAttr(
                hash, self.hash["NAME"] + " cmdIcon pause:rc_PAUSE play:rc_PLAY"
            )
            await fhem.CommandAttr(
                hash, self.hash["NAME"] + " webCmd volume:play:pause"
            )



import asyncio
from fhempy.lib.generic import FhemModule
from fhempy.lib import fhem, fhem_pythonbinding
from meross_iot.model.enums import OnlineStatus, Namespace


class meross_device:
    def __init__(self, logger, fhemdevice: FhemModule):
        self.logger = logger
        self.fhemdev = fhemdevice
        self.hash = fhemdevice.hash

    async def Define(self, hash, args, argsh):
        self._setupdev_name = args[3]
        self._deviceid = args[4]
        self._device = None
        self.hash["DEVICEID"] = self._deviceid

        self._setupdev = None
        await fhem.readingsSingleUpdate(self.hash, "state", "ready", 1)
        self.fhemdev.create_async_task(self._init_device())

    async def _get_set_commands(self):
        set_conf = {}
        set_conf["on"] = {}
        set_conf["off"] = {}
        self.fhemdev.set_set_config(set_conf)

    async def set_on(self, hash, params):
        await self._device.async_turn_on()

    async def set_off(self, hash, params):
        await self._device.async_turn_off()

    async def _init_device(self):
        try:
            await self._connect_to_setup_device()
            await self._setup_device()
            await self._get_set_commands()
            await self.update_readings()
        except Exception as ex:
            self.logger.exception(ex)

    async def _connect_to_setup_device(self):
        while self._setupdev is None or self._setupdev.ready is False:
            await asyncio.sleep(1)
            self._setupdev = fhem_pythonbinding.getFhemPyDeviceByName(
                self._setupdev_name
            )
            if self._setupdev:
                self._setupdev = self._setupdev.meross_device

    async def _async_push_notification_received(
        self, namespace: Namespace, data: dict, device_internal_id: str
    ):
        update_state = False
        full_update = False

        if namespace == Namespace.CONTROL_UNBIND:
            self.logger.warning(
                f"Received unbind event. Removing device {self._device.name} from FHEM"
            )
            await self.platform.async_remove_entity(self.entity_id)
        elif namespace == Namespace.SYSTEM_ONLINE:
            self.logger.warning(f"Device {self._device.name} reported online event.")
            online = OnlineStatus(int(data.get("online").get("status")))
            update_state = True
            full_update = online == OnlineStatus.ONLINE

        elif namespace == Namespace.HUB_ONLINE:
            self.logger.warning(
                f"Device {self._device.name} reported (HUB) online event."
            )
            online = OnlineStatus(int(data.get("status")))
            update_state = True
            full_update = online == OnlineStatus.ONLINE
        else:
            update_state = True
            full_update = False

        if full_update:
            await self._device.async_update()

        if update_state:
            await self.update_readings()

    async def _setup_device(self):
        self._device = self._setupdev.get_device_by_id(self._deviceid)

        if self._device is not None:
            await self._device.async_update()
            self._device.register_push_notification_handler_coroutine(
                self._async_push_notification_received
            )

    async def update_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "uuid", self._device.uuid)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "name", self._device.name)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "firmware_version", self._device.firmware_version
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "hardware_version", self._device.hardware_version
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "internal_id", self._device.internal_id
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "mqtt_host", self._device.mqtt_host
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "mqtt_port", self._device.mqtt_port
        )
        await fhem.readingsBulkUpdateIfChanged(self.hash, "type", self._device.type)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "online_status", self._device.online_status.name
        )
        onoff = "off"
        if self._device.is_on():
            onoff = "on"
        await fhem.readingsBulkUpdateIfChanged(self.hash, "state", onoff)
        await fhem.readingsEndUpdate(self.hash, 1)



import asyncio
import functools
import getopt
import importlib
import json
import logging
import site
import sys
import os
import time
import traceback

import websockets

from . import fhem, pkg_installer, utils, version
from .core.zeroconf import zeroconf

logger = logging.getLogger(__name__)

loadedModuleInstances = {}
moduleLoadingRunning = {}
zc_info = None

pip_lock = asyncio.Lock()

connection_start = 0
fct_timeout = 60

# internal modules
active_internal_modules = []
conf = {"internal_modules": ["discover_fhempy"]}


def getFhemPyDeviceByName(name):
    if name in loadedModuleInstances:
        return loadedModuleInstances[name]
    return None


async def activate_internal_modules():
    for im in conf["internal_modules"]:
        await pkg_installer.check_and_install_dependencies("core/" + im)
        module_object = importlib.import_module("fhempy.lib.core." + im + "." + im)
        # create instance of class with logger
        module_class = getattr(module_object, im)
        instance = module_class()
        await instance.activate()
        active_internal_modules.append(instance)


async def pybinding(websocket, path):
    global zc_info
    if zc_info is not None:
        # FHEM discovered us, stop zeroconf
        await zeroconf.get_instance(logger).unregister_service(zc_info)
        zc_info = None
        await zeroconf.get_instance(logger).stop()

    global connection_start
    connection_start = time.time()
    logger.info("Incoming FHEM connection: " + websocket.remote_address[0])
    pb = PyBinding(websocket)
    fhem.updateConnection(pb)
    await activate_internal_modules()
    await fhem.send_version()
    try:
        async for message in websocket:
            asyncio.create_task(pb.onMessage(message))
    except websockets.exceptions.ConnectionClosedError:
        logger.error("Connection closed error", exc_info=True)
        logger.info("Restart binding")
        sys.exit(1)


class PyBinding:

    msg_listeners = []

    def __init__(self, websocket):
        self.wsconnection = websocket

    def registerMsgListener(self, listener, awaitid):
        self.msg_listeners.append({"func": listener, "awaitId": awaitid})

    async def send(self, msg):
        await self.wsconnection.send(msg.encode("utf-8"))

    async def sendBackReturn(self, hash, ret):
        retHash = hash.copy()
        retHash["finished"] = 1
        retHash["returnval"] = ret
        retHash["id"] = hash["id"]
        msg = json.dumps(retHash)
        logger.debug("<<< WS: " + msg)
        await self.wsconnection.send(msg.encode("utf-8"))
        fhem.setFunctionInactive(hash)

    async def sendBackError(self, hash, error):
        logger.error(error + "(id: {})".format(hash["id"]))
        retHash = hash.copy()
        retHash["finished"] = 1
        retHash["error"] = error
        retHash["id"] = hash["id"]
        msg = json.dumps(retHash, ensure_ascii=False)
        logger.debug("<<< WS: " + msg)
        await self.wsconnection.send(msg.encode("utf-8"))
        fhem.setFunctionInactive(hash)

    async def updateHash(self, hash):
        retHash = hash.copy()
        retHash["msgtype"] = "update_hash"
        del retHash["id"]
        msg = json.dumps(retHash, ensure_ascii=False)
        logger.debug("<<< WS: " + msg)
        await self.wsconnection.send(msg.encode("utf-8"))

    def getLogLevel(self, verbose_level):
        if verbose_level == "5":
            return logging.DEBUG
        elif verbose_level == "4":
            return logging.INFO
        elif verbose_level == "3":
            return logging.WARNING
        else:
            return logging.ERROR

    async def onMessage(self, payload):
        try:
            await self._onMessage(payload)
        except SystemExit:
            sys.exit(1)
        except Exception:
            logger.exception("Failed to handle message: " + str(payload))

    async def _onMessage(self, payload):
        msg = payload
        logger.debug(">>> WS: " + msg)
        hash = None
        try:
            hash = json.loads(msg)
        except Exception:
            logger.error("Websocket JSON couldn't be decoded")
            return

        global fct_timeout, connection_start
        if time.time() - connection_start > 120:
            fct_timeout = 10

        try:
            if "awaitId" in hash and len(self.msg_listeners) > 0:
                removeElement = None
                for listener in self.msg_listeners:
                    if listener["awaitId"] == hash["awaitId"]:
                        listener["func"](msg)
                        removeElement = listener
                if removeElement:
                    self.msg_listeners.remove(removeElement)
            else:
                ret = ""
                if hash["msgtype"] == "update":
                    await fhem.readingsSingleUpdate(
                        hash, "version", "update started...", 1
                    )
                    await pkg_installer.force_update_package("fhempy")
                    await fhem.readingsSingleUpdate(
                        hash,
                        "version",
                        "update finished...please wait",
                        1,
                    )
                    os._exit(1)
                if hash["msgtype"] == "function":
                    # this is needed to avoid 2 replies on dep installation
                    fhem_reply_done = False
                    fhem.setFunctionActive(hash)
                    # load module
                    nmInstance = None
                    if hash["function"] == "Rename":
                        if hash["NAME"] in loadedModuleInstances:
                            loadedModuleInstances[
                                hash["args"][1]
                            ] = loadedModuleInstances[hash["args"][0]]
                            del loadedModuleInstances[hash["args"][0]]
                            await self.sendBackReturn(hash, "")
                            return 0

                    if hash["function"] != "Undefine":
                        # Load module and execute Define
                        # if Define isn't called right now
                        if not (hash["NAME"] in loadedModuleInstances):
                            if hash["NAME"] in moduleLoadingRunning:
                                await self.sendBackReturn(hash, "")
                                return 0

                            moduleLoadingRunning[hash["NAME"]] = True

                            # loading a module might take some time,
                            # therefore sendBackReturn now
                            await self.sendBackReturn(hash, "")
                            fhem_reply_done = True

                            try:
                                # check dependencies
                                deps_ok = pkg_installer.check_dependencies(
                                    hash["PYTHONTYPE"]
                                )
                                if deps_ok is False:
                                    # readingsSingleUpdate inform about dep installation
                                    await fhem.readingsSingleUpdate(
                                        hash, "state", "Installing updates...", 1
                                    )
                                    # run only one installation and do depcheck
                                    # before any other installation
                                    async with pip_lock:
                                        # make sure that all import caches
                                        # are up2date before check
                                        importlib.invalidate_caches()
                                        # check again if something
                                        # changed for dependencies
                                        deps_ok = pkg_installer.check_dependencies(
                                            hash["PYTHONTYPE"]
                                        )
                                        if deps_ok is False:
                                            # start installation in
                                            # a separate asyncio thread
                                            await pkg_installer.check_and_install_dependencies(
                                                hash["PYTHONTYPE"]
                                            )
                                            # update cache again after install
                                            if (
                                                not site.getusersitepackages()
                                                in sys.path
                                            ):
                                                logger.debug(
                                                    "add pip path: "
                                                    + site.getusersitepackages()
                                                )
                                                sys.path.append(
                                                    site.getusersitepackages()
                                                )
                                            importlib.invalidate_caches()
                                    # when installation finished, inform user
                                    await fhem.readingsSingleUpdate(
                                        hash,
                                        "state",
                                        "Installation finished. Please wait...",
                                        1,
                                    )
                                    # wait 3s so that user can read
                                    # the msg about installation
                                    await asyncio.sleep(3)
                                    # continue define

                                # import module
                                pymodule = (
                                    "fhempy.lib."
                                    + hash["PYTHONTYPE"]
                                    + "."
                                    + hash["PYTHONTYPE"]
                                )
                                # import might take a long time, therefore run_blocking
                                module_object = await utils.run_blocking(
                                    functools.partial(importlib.import_module, pymodule)
                                )
                                # create instance of class with logger
                                target_class = getattr(
                                    module_object, hash["PYTHONTYPE"]
                                )
                                moduleLogger = logging.getLogger(hash["NAME"])
                                moduleLogger.setLevel(
                                    self.getLogLevel(
                                        await fhem.AttrVal(hash["NAME"], "verbose", "3")
                                    )
                                )
                                loadedModuleInstances[hash["NAME"]] = target_class(
                                    moduleLogger
                                )
                                del moduleLoadingRunning[hash["NAME"]]
                                if hash["function"] != "Define":
                                    func = getattr(
                                        loadedModuleInstances[hash["NAME"]],
                                        "Define",
                                        "nofunction",
                                    )
                                    await asyncio.wait_for(
                                        func(hash, hash["defargs"], hash["defargsh"]),
                                        fct_timeout,
                                    )
                            except asyncio.TimeoutError:
                                errorMsg = (
                                    f"Function execution >{fct_timeout}s, "
                                    f"cancelled: {hash['NAME']} Define"
                                )
                                if fhem_reply_done:
                                    await fhem.readingsSingleUpdate(
                                        hash, "state", errorMsg, 1
                                    )
                                else:
                                    await self.sendBackError(hash, errorMsg)
                                return 0
                            except ModuleNotFoundError:
                                errorMsg = (
                                    f"Module failed to load: {hash['PYTHONTYPE']}\n"
                                    f"Maybe you need to update fhempy "
                                    f"on this or remote peer."
                                )
                                errorMsg += "\n\nStacktrace:\n" + traceback.format_exc()
                                if fhem_reply_done:
                                    await fhem.readingsSingleUpdate(
                                        hash, "state", errorMsg, 1
                                    )
                                else:
                                    await self.sendBackError(hash, errorMsg)
                                return 0
                            except Exception:
                                errorMsg = (
                                    "Failed to load module "
                                    + hash["PYTHONTYPE"]
                                    + ": "
                                    + traceback.format_exc()
                                )
                                if fhem_reply_done:
                                    await fhem.readingsSingleUpdate(
                                        hash, "state", errorMsg, 1
                                    )
                                else:
                                    await self.sendBackError(hash, errorMsg)
                                return 0

                    try:
                        nmInstance = loadedModuleInstances[hash["NAME"]]
                    except Exception:
                        if hash["function"] != "Undefine":
                            logging.getLogger(hash["NAME"]).exception(
                                f"Couldn't handle {msg}"
                            )
                        nmInstance = None

                    if nmInstance is not None:
                        try:
                            # handle verbose level of logging
                            if (
                                hash["function"] == "Attr"
                                and hash["args"][2] == "verbose"
                            ):
                                moduleLogger = logging.getLogger(hash["NAME"])
                                if hash["args"][0] == "set":
                                    moduleLogger.setLevel(
                                        self.getLogLevel(hash["args"][3])
                                    )
                                else:
                                    moduleLogger.setLevel(logging.ERROR)
                            else:
                                # call Set/Attr/Define/...
                                func = getattr(
                                    nmInstance, hash["function"], "nofunction"
                                )
                                if func != "nofunction":
                                    logger.debug(
                                        (
                                            f"Start function {hash['NAME']}:"
                                            f"{hash['function']}"
                                        )
                                    )
                                    if hash["function"] == "Undefine":
                                        try:
                                            ret = await asyncio.wait_for(
                                                func(hash), fct_timeout
                                            )
                                        except Exception:
                                            del loadedModuleInstances[hash["NAME"]]
                                    else:
                                        ret = await asyncio.wait_for(
                                            func(hash, hash["args"], hash["argsh"]),
                                            fct_timeout,
                                        )
                                    logger.debug(
                                        (
                                            f"End function {hash['NAME']}:"
                                            f"{hash['function']}"
                                        )
                                    )
                                    if ret is None:
                                        ret = ""
                                    if fhem_reply_done:
                                        await self.updateHash(hash)
                        except asyncio.TimeoutError:
                            errorMsg = (
                                f"Function execution >{fct_timeout}s, "
                                f"cancelled: {hash['NAME']} - {hash['function']}"
                            )
                            if fhem_reply_done:
                                await fhem.readingsSingleUpdate(
                                    hash, "state", errorMsg, 1
                                )
                            else:
                                await self.sendBackError(hash, errorMsg)
                            return 0
                        except Exception:
                            errorMsg = (
                                "Failed to execute function "
                                + hash["function"]
                                + ": "
                                + traceback.format_exc()
                            )
                            if fhem_reply_done:
                                await fhem.readingsSingleUpdate(
                                    hash, "state", errorMsg, 1
                                )
                            else:
                                await self.sendBackError(hash, errorMsg)
                            return 0

                    if hash["function"] == "Undefine":
                        if hash["NAME"] in loadedModuleInstances:
                            del loadedModuleInstances[hash["NAME"]]

                    if fhem_reply_done is False:
                        await self.sendBackReturn(hash, ret)

        except SystemExit as se:
            raise se
        except Exception:
            logger.error("Failed to handle message: ", exc_info=True)


def usage():
    print("Usage: fhempy [-h|--help] [-v|--version] [-i|--ip] [-p|--port] [-l|--local]")
    print("  --local   Use only if you run fhempy on your FHEM machine")
    print(
        "  --ip      "
        "Specify the IP address for FHEM connection setup (default: local ip)"
    )
    print("  --port    Specify the port fhempy runs on (default: 15733)")
    print("  --version Print version and exit")
    print("  --help    This help text")


async def async_main():
    try:
        opts, args = getopt.getopt(
            sys.argv[1:],
            "dhvli:p:",
            ["help", "version", "ip=", "port=", "local", "debug"],
        )
    except getopt.GetoptError as err:
        logger.error(err)
        usage()
        sys.exit(2)

    ip = None
    port = 15733
    local = False
    for o, a in opts:
        if o in ("-v", "--version"):
            print("fhempy " + str(version.__version__))
            sys.exit()
        elif o in ("-h", "--help"):
            usage()
            sys.exit()
        elif o in ("-i", "--ip"):
            ip = a
        elif o in ("-p", "--port"):
            port = a
        elif o in ("-l", "--local"):
            local = True
        elif o in ("-d", "--debug"):
            logging.getLogger("").setLevel(logging.DEBUG)

    logger.info("Starting fhempy...")

    await pkg_installer.check_and_install_dependencies("core")

    if not local:
        logger.info("Advertise fhempy on local network")
        # running on remote peer, start zeroconf for autodiscovery
        if ip is None:
            local_ip = utils.get_local_ip()
        else:
            local_ip = ip
        zc = zeroconf.get_instance(logger)
        global zc_info
        zc_info = await zc.register_service(
            "_http",
            "fhempy (" + local_ip + ")",
            port,
            {"port": port, "ip": local_ip},
        )

    logger.info("Waiting for FHEM connection")
    await websockets.serve(
        pybinding, "0.0.0.0", port, ping_timeout=None, ping_interval=None
    )


def run():
    logging.getLogger("asyncio").setLevel(logging.WARNING)
    asyncio.get_event_loop().set_debug(True)
    asyncio.get_event_loop().run_until_complete(async_main())
    asyncio.get_event_loop().run_forever()



# Copyright 2020 Elelabs International Limited

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import binascii
import io
import socket
import time

from xmodem import XMODEM


# Maximum untouched utility with fix only serial class from pyserial to TCP
# https://github.com/Elelabs/elelabs-zigbee-ezsp-utility
class serial:
    PARITY_NONE = None
    STOPBITS_ONE = None

    class Serial:
        def __init__(self, port, **kwargs):
            self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.s.settimeout(5)
            self.s.connect(port)

        def flushInput(self):
            pass

        def read(self, size: int = 1):
            try:
                return self.s.recv(size)
            except:
                return b""

        def readline(self):
            raw = b""
            while True:
                c = self.read()
                raw += c
                if c == b"\n" or c == b"":
                    break
            return raw

        def write(self, data: bytes):
            self.s.send(data)

        def close(self):
            self.s.close()


class AdapterModeProbeStatus:
    NORMAL = 0
    BOOTLOADER = 1
    ERROR = 2


class SerialInterface:
    def __init__(self, port, baudrate):
        self.port = port
        self.baudrate = baudrate

    def open(self):
        try:
            self.serial = serial.Serial(
                port=self.port,
                baudrate=self.baudrate,
                parity=serial.PARITY_NONE,
                stopbits=serial.STOPBITS_ONE,
                xonxoff=True,
                timeout=3,
            )
        except Exception as e:
            raise Exception("PORT ERROR: %s" % str(e))

    def close(self):
        self.serial.close()


class AshProtocolInterface:
    FLAG_BYTE = b"\x7E"
    RANDOMIZE_START = 0x42
    RANDOMIZE_SEQ = 0xB8
    RSTACK_FRAME_CMD = b"\x1A\xC0\x38\xBC\x7E"
    RSTACK_FRAME_ACK = b"\x1A\xC1\x02\x0B\x0A\x52\x7E"

    def __init__(self, serial, config, logger):
        self.logger = logger
        self.config = config
        self.serial = serial

        self.ackNum = 0
        self.frmNum = 0

    def dataRandomize(self, frame):
        rand = self.RANDOMIZE_START
        out = bytearray()
        for x in frame:
            out += bytearray([x ^ rand])
            if rand % 2:
                rand = (rand >> 1) ^ self.RANDOMIZE_SEQ
            else:
                rand = rand >> 1
        return out

    def ashFrameBuilder(self, ezsp_frame):
        ash_frame = bytearray()
        # Control byte
        ash_frame += bytearray(
            [(((self.ackNum << 0) & 0xFF) | (((self.frmNum % 8) << 4) & 0xFF)) & 0xFF]
        )
        self.ackNum = (self.ackNum + 1) % 8
        self.frmNum = (self.frmNum + 1) % 8
        ash_frame += self.dataRandomize(ezsp_frame)
        crc = binascii.crc_hqx(ash_frame, 0xFFFF)
        ash_frame += bytearray([crc >> 8, crc & 0xFF])
        ash_frame = self.replaceReservedBytes(ash_frame)
        ash_frame += self.FLAG_BYTE
        if self.config.dlevel == "ASH":
            self.logger.debug(
                "[ ASH  REQUEST ] " + " ".join(format(x, "02x") for x in ash_frame)
            )
        return ash_frame

    def revertEscapedBytes(self, msg):
        msg = msg.replace(b"\x7d\x5d", b"\x7d")
        msg = msg.replace(b"\x7d\x5e", b"\x7e")
        msg = msg.replace(b"\x7d\x31", b"\x11")
        msg = msg.replace(b"\x7d\x33", b"\x13")
        msg = msg.replace(b"\x7d\x38", b"\x18")
        msg = msg.replace(b"\x7d\x3a", b"\x1a")
        return msg

    def replaceReservedBytes(self, msg):
        msg = msg.replace(b"\x7d", b"\x7d\x5d")
        msg = msg.replace(b"\x7e", b"\x7d\x5e")
        msg = msg.replace(b"\x11", b"\x7d\x31")
        msg = msg.replace(b"\x13", b"\x7d\x33")
        msg = msg.replace(b"\x18", b"\x7d\x38")
        msg = msg.replace(b"\x1a", b"\x7d\x3a")
        return msg

    def getResponse(self, applyRandomize=False):
        timeout = time.time() + 3
        msg = bytearray()

        receivedbyte = None

        while (time.time() < timeout) and (receivedbyte != self.FLAG_BYTE):
            receivedbyte = self.serial.read()
            msg += receivedbyte

        if len(msg) == 0:
            return -1, None, None

        msg = self.revertEscapedBytes(msg)

        if self.config.dlevel == "ASH":
            self.logger.debug(
                "[ ASH RESPONSE ] " + " ".join(format(x, "02x") for x in msg)
            )

        if applyRandomize:
            msg_parsed = self.dataRandomize(bytearray(msg[1:-3]))
            if self.config.dlevel == "ASH" or self.config.dlevel == "EZSP":
                self.logger.debug(
                    "[ EZSP RESPONSE ] "
                    + " ".join(format(x, "02x") for x in msg_parsed)
                )
            return 0, msg, msg_parsed
        else:
            return 0, msg

    def sendResetFrame(self):
        self.serial.flushInput()
        self.logger.debug("RESET FRAME")
        if self.config.dlevel == "ASH":
            self.logger.debug(
                "[ ASH  REQUEST ] "
                + " ".join(format(x, "02x") for x in self.RSTACK_FRAME_CMD)
            )
        self.serial.write(self.RSTACK_FRAME_CMD)
        status, response = self.getResponse()

        if status:
            return status

        if not (self.RSTACK_FRAME_ACK in response):
            return -1

        return 0

    def sendAck(self, ackNum):
        ack = bytearray([ackNum & 0x07 | 0x80])
        crc = binascii.crc_hqx(ack, 0xFFFF)
        ack += bytearray([crc >> 8, crc & 0xFF])
        ack = self.replaceReservedBytes(ack)
        ack += self.FLAG_BYTE

        if self.config.dlevel == "ASH":
            self.logger.debug("[ ASH ACK ] " + " ".join(format(x, "02x") for x in ack))
        self.serial.write(ack)

    def sendAshCommand(self, ezspFrame):
        ash_frame = self.ashFrameBuilder(ezspFrame)
        self.serial.flushInput()
        self.serial.write(ash_frame)
        status, ash_response, ezsp_response = self.getResponse(True)
        if status:
            return status, None

        self.sendAck(ash_response[0])
        return 0, ezsp_response


class EzspProtocolInterface:
    def __init__(self, serial, config, logger):
        self.logger = logger
        self.config = config

        self.INITIAL_EZSP_VERSION = 4

        self.VERSION = b"\x00"
        self.GET_VALUE = b"\xAA"
        self.GET_MFG_TOKEN = b"\x0B"
        self.LAUNCH_STANDALONE_BOOTLOADER = b"\x8F"

        self.EZSP_VALUE_VERSION_INFO = 0x11
        self.EZSP_MFG_STRING = 0x01
        self.EZSP_MFG_BOARD_NAME = 0x02
        self.STANDALONE_BOOTLOADER_NORMAL_MODE = 1

        self.ezspVersion = self.INITIAL_EZSP_VERSION
        self.sequenceNum = 0
        self.ash = AshProtocolInterface(serial, config, logger)

    def ezspFrameBuilder(self, command):
        ezsp_frame = bytearray()

        # Sequence byte
        ezsp_frame += bytearray([self.sequenceNum])
        self.sequenceNum = (self.sequenceNum + 1) % 255
        ezsp_frame += b"\x00"
        if self.ezspVersion >= 5:
            # Legacy frame ID - always 0xFF
            ezsp_frame += b"\xFF"
            # Extended frame control
            ezsp_frame += b"\x00"

        ezsp_frame = ezsp_frame + command

        if self.ezspVersion >= 8:
            ezsp_frame[2] = 0x01
            ezsp_frame[3] = command[0] & 0xFF  # LSB
            ezsp_frame[4] = command[0] >> 8  # MSB

        if self.config.dlevel == "ASH" or self.config.dlevel == "EZSP":
            self.logger.debug(
                "[ EZSP  REQUEST ] " + " ".join(format(x, "02x") for x in ezsp_frame)
            )
        return ezsp_frame

    def sendEzspCommand(self, commandData, commandName=""):
        self.logger.debug(commandName)
        status, response = self.ash.sendAshCommand(self.ezspFrameBuilder(commandData))
        if status:
            raise Exception("sendAshCommand status error: %d" % status)

        return response

    def sendVersion(self, desiredProtocolVersion):
        resp = self.sendEzspCommand(
            self.VERSION + bytearray([desiredProtocolVersion]),
            "sendVersion: V%d" % desiredProtocolVersion,
        )
        return resp[3]  # protocolVersion

    def getValue(self, valueId, valueIdName):
        resp = self.sendEzspCommand(
            self.GET_VALUE + bytearray([valueId]), "getValue: %s" % valueIdName
        )
        status = resp[5]
        valueLength = resp[6]
        valueArray = resp[7:]
        return status, valueLength, valueArray

    def getMfgToken(self, tokenId, tokenIdName):
        resp = self.sendEzspCommand(
            self.GET_MFG_TOKEN + bytearray([tokenId]), "getMfgToken: %s" % tokenIdName
        )
        tokenDataLength = resp[5]
        tokenData = resp[6:]
        return tokenDataLength, tokenData

    def launchStandaloneBootloader(self, mode, modeName):
        resp = self.sendEzspCommand(
            self.LAUNCH_STANDALONE_BOOTLOADER + bytearray([mode]),
            "launchStandaloneBootloader: %s" % modeName,
        )
        status = resp[5]
        return status

    def initEzspProtocol(self):
        ash_status = self.ash.sendResetFrame()
        if ash_status:
            return ash_status

        self.ezspVersion = self.sendVersion(self.INITIAL_EZSP_VERSION)
        self.logger.debug("EZSP v%d detected" % self.ezspVersion)
        if self.ezspVersion != self.INITIAL_EZSP_VERSION:
            self.sendVersion(self.ezspVersion)

        return 0


class ElelabsUtilities:
    def __init__(self, config, logger):
        self.logger = logger
        self.config = config

    def probe(self):
        serialInterface = SerialInterface(self.config.port, self.config.baudrate)
        serialInterface.open()

        ezsp = EzspProtocolInterface(serialInterface.serial, self.config, self.logger)
        ezsp_status = ezsp.initEzspProtocol()
        if ezsp_status == 0:
            status, value_length, value_array = ezsp.getValue(
                ezsp.EZSP_VALUE_VERSION_INFO, "EZSP_VALUE_VERSION_INFO"
            )
            if status == 0:
                firmware_version = (
                    str(value_array[2])
                    + "."
                    + str(value_array[3])
                    + "."
                    + str(value_array[4])
                    + "-"
                    + str(value_array[0])
                )
            else:
                self.logger.info("EZSP status returned %d" % status)

            token_data_length, token_data = ezsp.getMfgToken(
                ezsp.EZSP_MFG_STRING, "EZSP_MFG_STRING"
            )
            if token_data.decode("ascii", "ignore") == "Elelabs":
                token_data_length, token_data = ezsp.getMfgToken(
                    ezsp.EZSP_MFG_BOARD_NAME, "EZSP_MFG_BOARD_NAME"
                )
                adapter_name = token_data.decode("ascii", "ignore")

                self.logger.info("Elelabs adapter detected:")
                self.logger.info("Adapter: %s" % adapter_name)
            else:
                adapter_name = None
                self.logger.info("Generic EZSP adapter detected:")

            self.logger.info("Firmware: %s" % firmware_version)
            self.logger.info("EZSP v%d" % ezsp.ezspVersion)

            serialInterface.close()
            return (
                AdapterModeProbeStatus.NORMAL,
                ezsp.ezspVersion,
                firmware_version,
                adapter_name,
            )
        else:
            if self.config.baudrate != 115200:
                serialInterface.close()
                time.sleep(1)
                serialInterface = SerialInterface(self.config.port, 115200)
                serialInterface.open()

            # check if allready in bootloader mode
            serialInterface.serial.write(b"\x0A")
            first_line = serialInterface.serial.readline()  # read blank line
            if len(first_line) == 0:
                # timeout
                serialInterface.close()
                self.logger.info(
                    "Couldn't communicate with the adapter in normal or in bootloader modes"
                )
                return AdapterModeProbeStatus.ERROR, None, None, None

            btl_info = (
                serialInterface.serial.readline()
            )  # read Gecko BTL version or blank line

            self.logger.info("EZSP adapter in bootloader mode detected:")
            self.logger.info(
                btl_info.decode("ascii", "ignore")[:-2]
            )  # show Bootloader version
            serialInterface.close()
            return AdapterModeProbeStatus.BOOTLOADER, None, None, None

    def restart(self, mode):
        adapter_status, ezsp_version, firmware_version, adapter_name = self.probe()
        if adapter_status == AdapterModeProbeStatus.NORMAL:
            if mode == "btl":
                serialInterface = SerialInterface(
                    self.config.port, self.config.baudrate
                )
                serialInterface.open()

                self.logger.info("Launch in bootloader mode")
                ezsp = EzspProtocolInterface(
                    serialInterface.serial, self.config, self.logger
                )
                ezsp_status = ezsp.initEzspProtocol()
                status = ezsp.launchStandaloneBootloader(
                    ezsp.STANDALONE_BOOTLOADER_NORMAL_MODE,
                    "STANDALONE_BOOTLOADER_NORMAL_MODE",
                )
                if status:
                    serialInterface.close()
                    self.logger.critical(
                        "Error launching the adapter in bootloader mode"
                    )
                    return -1

                serialInterface.close()
                # wait for reboot
                time.sleep(2)

                (
                    adapter_status,
                    ezsp_version,
                    firmware_version,
                    adapter_name,
                ) = self.probe()
                if adapter_status == AdapterModeProbeStatus.BOOTLOADER:
                    return 0
                else:
                    return -1
            else:
                self.logger.info("Allready in EZSP normal mode. No need to restart")
                return 0
        elif adapter_status == AdapterModeProbeStatus.BOOTLOADER:
            if mode == "btl":
                self.logger.info("Allready in bootloader mode. No need to restart")
                return 0
            else:
                serialInterface = SerialInterface(self.config.port, 115200)
                serialInterface.open()

                self.logger.info("Launch in EZSP normal mode")

                # Send Reboot
                serialInterface.serial.write(b"2")
                serialInterface.close()

                # wait for reboot
                time.sleep(2)

                (
                    adapter_status,
                    ezsp_version,
                    firmware_version,
                    adapter_name,
                ) = self.probe()
                if adapter_status == AdapterModeProbeStatus.NORMAL:
                    return 0
                else:
                    return -1

    def flash(self, filename):
        # STATIC FUNCTIONS
        def getc(size, timeout=1):
            read_data = self.serialInterface.serial.read(size)
            return read_data

        def putc(data, timeout=1):
            self.currentPacket += 1
            if (self.currentPacket % 20) == 0:
                print(".", end="")
            if (self.currentPacket % 100) == 0:
                print("")
            self.serialInterface.serial.write(data)
            time.sleep(0.001)

        # if not (".gbl" in filename) and not (".ebl" in filename):
        #     self.logger.critical(
        #         'Aborted! Gecko bootloader accepts .gbl or .ebl images only.')
        #     return

        if self.restart("btl"):
            self.logger.critical(
                "EZSP adapter not in the bootloader mode. Can't perform update procedure"
            )

        self.serialInterface = SerialInterface(self.config.port, 115200)
        self.serialInterface.open()
        # Enter '1' to initialize X-MODEM mode
        self.serialInterface.serial.write(b"\x0A")
        self.serialInterface.serial.write(b"1")
        time.sleep(1)
        self.serialInterface.serial.readline()  # BL > 1
        self.serialInterface.serial.readline()  # begin upload

        self.logger.info(
            "Successfully restarted into X-MODEM mode! Starting upload of the new firmware... DO NOT INTERRUPT(!)"
        )

        self.currentPacket = 0
        # Wait for char 'C'
        success = False
        start_time = time.time()
        while time.time() - start_time < 10:
            if self.serialInterface.serial.read() == b"C":
                success = True
                if time.time() - start_time > 5:
                    break
        if not success:
            self.logger.info(
                "Failed to restart into bootloader mode. Please see users guide."
            )
            return

        # Start XMODEM transaction
        modem = XMODEM(getc, putc)
        # stream = open(filename, 'rb')
        stream = io.BytesIO(filename)
        sentcheck = modem.send(stream)

        print("")
        if sentcheck:
            self.logger.info("Firmware upload complete")
        else:
            self.logger.critical(
                "Firmware upload failed. Please try a correct firmware image or restart in normal mode."
            )
            return
        self.logger.info("Rebooting NCP...")
        # Wait for restart
        time.sleep(4)
        # Send Reboot into App-Code command
        self.serialInterface.serial.write(b"2")
        self.serialInterface.close()
        time.sleep(2)
        return self.probe()

    def ele_update(self, new_version):
        adapter_status, ezsp_version, firmware_version, adapter_name = self.probe()
        if adapter_status == AdapterModeProbeStatus.NORMAL:
            if adapter_name == None:
                self.logger.critical(
                    "No Elelabs product detected.\r\nUse 'flash' utility for generic EZSP products.\r\nContact info@elelabs.com if you see this meesage for original Elelabs product"
                )
                return

            if new_version == "v6" and ezsp_version == 6:
                self.logger.info(
                    "Elelabs product is operating EZSP protocol v%d. No need to update to %s"
                    % (ezsp_version, new_version)
                )
                return

            if new_version == "v8" and ezsp_version == 8:
                self.logger.info(
                    "Elelabs product is operating EZSP protocol v%d. No need to update to %s"
                    % (ezsp_version, new_version)
                )
                return

            if adapter_name == "ELR023" or adapter_name == "ELU013":
                if new_version == "v6":
                    self.flash("data/ELX0X3_MG13_6.0.3_ezsp_v6.gbl")
                elif new_version == "v8":
                    self.flash("data/ELX0X3_MG13_6.7.0_ezsp_v8.gbl")
                else:
                    self.logger.critical("Unknown EZSP version")
            elif adapter_name == "ELR022" or adapter_name == "ELU012":
                self.logger.critical("TODO!. Contact Elelabs at info@elelabs.com")
            elif adapter_name == "EZBPIS" or adapter_name == "EZBUSBA":
                self.logger.critical("TODO!. Contact Elelabs at info@elelabs.com")
            else:
                self.logger.critical(
                    "Unknown Elelabs product %s detected.\r\nContact info@elelabs.com if you see this meesage for original Elelabs product"
                    % adapter_name
                )
        elif adapter_status == AdapterModeProbeStatus.BOOTLOADER:
            self.logger.critical(
                "The product not in the normal EZSP mode.\r\n'restart' into normal mode or use 'flash' utility instead"
            )
        else:
            self.logger.critical("No upgradable device found")



import asyncio
import socket

from zeroconf.asyncio import AsyncServiceInfo, AsyncZeroconf


class zeroconf:

    instance = None

    @staticmethod
    def get_instance(logger):
        if zeroconf.instance is None:
            zeroconf.instance = zeroconf(logger)
        return zeroconf.instance

    def __init__(self, logger):
        self.logger = logger
        self.loop = asyncio.get_event_loop()
        self.zeroconf = AsyncZeroconf()

    async def register_service(self, type, name, port, properties):
        hostname = socket.gethostname()
        local_ip = socket.gethostbyname(hostname)
        info = AsyncServiceInfo(
            type + "._tcp.local.",
            name + "." + type + "._tcp.local.",
            addresses=[socket.inet_aton(local_ip)],
            port=port,
            properties=properties,
            server=hostname + ".local.",
        )
        await self.zeroconf.async_register_service(info)
        return info

    async def unregister_service(self, info):
        await self.zeroconf.async_unregister_service(info)

    def get_zeroconf(self):
        return self.zeroconf

    async def stop(self):
        await self.zeroconf.async_close()



import asyncio
import concurrent.futures
import json
import logging
import random
import traceback

import websockets

from .version import __version__

logger = logging.getLogger(__name__)
logger.setLevel(logging.ERROR)

function_active = []
update_locks = {}
wsconnection = None

# TODO use run_coroutine_threadsafe if asyncio.get_event_loop() == None
# this would make all functions threadsafe


def updateConnection(ws):
    global wsconnection
    wsconnection = ws


def setFunctionActive(hash):
    function_active.append(hash["NAME"])


def setFunctionInactive(hash):
    element = function_active.pop()
    if element != hash["NAME"]:
        logger.error(
            f"Set wrong function inactive, tried {hash['NAME']}, "
            f"current function_active: {function_active},{element}"
        )


async def getDeviceHashName(hash, typeinternal, typevalue, internal, value):
    cmd = (
        "foreach my $fhem_dev (sort keys %main::defs) {"
        + "  return $main::defs($fhem_dev}{NAME}) if(defined($main::defs{$fhem_dev}{"
        + typeinternal
        + "}) && $main::defs{$fhem_dev}{"
        + typeinternal
        + "} eq '"
        + typevalue
        + "' && $main::defs{$fhem_dev}{"
        + internal
        + "} eq '"
        + value
        + "');;"
        + "}"
        + "return 0;;"
    )
    return await sendCommandHash(hash, cmd)


async def getUniqueId(hash):
    cmd = "getUniqueId()"
    return await sendCommandHash(hash, cmd)


async def init_done(hash):
    cmd = "$init_done"
    return await sendCommandHash(hash, cmd)


async def ReadingsVal(name, reading, default):
    cmd = "ReadingsVal('" + name + "', '" + reading + "', '" + default + "')"
    return await sendCommandName(name, cmd)


async def AttrVal(name, attr, default):
    cmd = "AttrVal('" + name + "', '" + attr + "', '" + default + "')"
    return await sendCommandName(name, cmd)


async def InternalVal(name, internal, default):
    cmd = "InternalVal('" + name + "', '" + internal + "', '" + default + "')"
    return await sendCommandName(name, cmd)


async def addToDevAttrList(name, attr_list):
    cmd = "addToDevAttrList('" + name + "', '" + attr_list + "')"
    return await sendCommandName(name, cmd)


async def setDevAttrList(name, attr_list):
    attr_list += " IODev"
    cmd = "setDevAttrList('" + name + "', '" + attr_list + " '.$readingFnAttributes)"
    return await sendCommandName(name, cmd)


async def readingsBeginUpdate(hash):
    if hash["NAME"] not in update_locks:
        update_locks[hash["NAME"]] = asyncio.Lock()
    await update_locks[hash["NAME"]].acquire()
    cmd = "readingsBeginUpdate($defs{'" + hash["NAME"] + "'});;"
    return await sendCommandHash(hash, cmd)


async def readingsBulkUpdateIfChanged(hash, reading, value):
    try:
        value = convertValue(value)
        cmd = (
            "readingsBulkUpdateIfChanged($defs{'"
            + hash["NAME"]
            + "'},'"
            + reading
            + "','"
            + value.replace("'", "\\'")
            + "');;"
        )
        return await sendCommandHash(hash, cmd)
    except Exception:
        logger.exception("Failed to do readingsBulkUpdateIfChanged")


async def readingsBulkUpdate(hash, reading, value, changed=None):
    try:
        value = convertValue(value)
        if changed is None:
            cmd = (
                "readingsBulkUpdate($defs{'"
                + hash["NAME"]
                + "'},'"
                + reading
                + "','"
                + value.replace("'", "\\'")
                + "');;"
            )
        else:
            cmd = (
                "readingsBulkUpdate($defs{'"
                + hash["NAME"]
                + "'},'"
                + reading
                + "','"
                + value.replace("'", "\\'")
                + "', "
                + str(changed)
                + ");;"
            )
        return await sendCommandHash(hash, cmd)
    except Exception:
        logger.exception("Failed to do readingsBulkUpdate")


async def readingsEndUpdate(hash, do_trigger):
    cmd = "readingsEndUpdate($defs{'" + hash["NAME"] + "'}," + str(do_trigger) + ");;"
    res = await sendCommandHash(hash, cmd)
    update_locks[hash["NAME"]].release()
    return res


async def readingsSingleUpdate(hash, reading, value, do_trigger):
    if hash["NAME"] not in update_locks:
        update_locks[hash["NAME"]] = asyncio.Lock()
    async with update_locks[hash["NAME"]]:
        value = convertValue(value)
        cmd = (
            "readingsSingleUpdate($defs{'"
            + hash["NAME"]
            + "'},'"
            + reading
            + "','"
            + value.replace("'", "\\'")
            + "',"
            + str(do_trigger)
            + ")"
        )
        return await sendCommandHash(hash, cmd)


async def readingsSingleUpdateIfChanged(hash, reading, value, do_trigger):
    if hash["NAME"] not in update_locks:
        update_locks[hash["NAME"]] = asyncio.Lock()
    async with update_locks[hash["NAME"]]:
        value = convertValue(value)
        cmd = (
            "readingsBeginUpdate($defs{'"
            + hash["NAME"]
            + "'});;readingsBulkUpdateIfChanged($defs{'"
            + hash["NAME"]
            + "'},'"
            + reading
            + "','"
            + value.replace("'", "\\'")
            + "');;readingsEndUpdate($defs{'"
            + hash["NAME"]
            + "'},"
            + str(do_trigger)
            + ");;"
        )
        return await sendCommandHash(hash, cmd)


async def CommandDefine(hash, definition: str):
    cmd = 'CommandDefine(undef, "' + definition + '")'
    ret = await sendCommandHash(hash, cmd)
    if ret is not None:
        return ret

    pos_pythonmodule = definition.find("PythonModule")
    if pos_pythonmodule > 0:
        devname = definition.split(" ")[0]
        iodev = await AttrVal(hash["NAME"], "IODev", "")
        if iodev != "":
            await CommandAttr(hash, f"{devname} IODev {iodev}")

        fhempytype = definition.split(" ")[2]
        await CommandAttr(hash, f"{devname} group {fhempytype}")

        room = await AttrVal(hash["NAME"], "room", "")
        if room != "":
            await CommandAttr(hash, f"{devname} room fhempy")


async def CommandAttr(hash, attrdef):
    cmd = 'CommandAttr(undef, "' + attrdef.replace('"', '\\"') + '")'
    return await sendCommandHash(hash, cmd)


async def CommandDeleteReading(hash, deldef):
    cmd = 'CommandDeleteReading(undef, "' + deldef + '")'
    return await sendCommandHash(hash, cmd)


async def checkIfDeviceExists(hash, typeinternal, typevalue, internal, value):
    cmd = (
        "foreach my $fhem_dev (sort keys %main::defs) {"
        + "  return 1 if(defined($main::defs{$fhem_dev}{"
        + typeinternal
        + "}) && $main::defs{$fhem_dev}{"
        + typeinternal
        + "} eq '"
        + typevalue
        + "' && $main::defs{$fhem_dev}{"
        + internal
        + "} eq '"
        + value
        + "');;"
        + "}"
        + "return 0;;"
    )
    return await sendCommandHash(hash, cmd)


# UTILS FUNCTIONS TO SEND COMMAND TO FHEM
def convertValue(value):
    if value is None:
        value = ""
    if value is True:
        value = 1
    if value is False:
        value = 0

    return str(value)


async def send_version():
    msg = {
        "msgtype": "version",
        "version": __version__,
    }
    msg = json.dumps(msg, ensure_ascii=False)
    logger.debug("<<< WS: " + msg)
    global wsconnection
    await wsconnection.send(msg)


async def send_and_wait(name, cmd):
    fut = asyncio.get_running_loop().create_future()
    msg = {
        "awaitId": random.randint(10000000, 99999999),
        "NAME": name,
        "msgtype": "command",
        "command": cmd,
    }

    def listener(rmsg):
        try:
            fut.set_result(rmsg)
        except Exception:
            logger.error("Failed to set result, received: " + rmsg)

    global wsconnection
    wsconnection.registerMsgListener(listener, msg["awaitId"])
    msg = json.dumps(msg, ensure_ascii=False)
    logger.debug("<<< WS: " + msg)
    try:
        await wsconnection.send(msg)
        logger.debug("message sent successfully")
    except websockets.exceptions.ConnectionClosed:
        logger.error("Connection closed, can't send message.")
    except Exception as e:
        logger.error("Failed to send message via websocket: " + e)
        fut.set_exception(Exception("Failed to send message via websocket"))

    return await fut


async def sendCommandName(name, cmd, hash=None):
    ret = ""
    try:
        logger.debug("sendCommandName START")
        while len(function_active) != 0:
            if function_active[-1] == name:
                break
            await asyncio.sleep(0.001)
        # wait max 1s for reply from FHEM
        jsonmsg = await asyncio.wait_for(send_and_wait(name, cmd), 15)
        logger.debug("sendCommandName END")
        ret = json.loads(jsonmsg)["result"]
    except asyncio.TimeoutError:
        logger.error("Timeout - NO RESPONSE for command: " + cmd)
        ret = ""
    except concurrent.futures.CancelledError:
        # function timeout
        pass
    except Exception as e:
        logger.error("Exception while waiting for reply: " + e)
        traceback.format_exc()
        ret = str(e)

    return ret


async def sendCommandHash(hash, cmd):
    return await sendCommandName(hash["NAME"], cmd, hash)



"""
Logging can be setup from:

1. Hass default config

```yaml
logger:
 logs:
   custom_components.xiaomi_gateway3: debug
```

2. Integration config (YAML)

```yaml
xiaomi_gateway3:
  logger:
    filename: xiaomi_gateway3.log
    propagate: False  # disable log to home-assistant.log and console
    max_bytes: 100000000
    backup_count: 3
```

3. Integration config (GUI)

Configuration > Xiaomi Gateway 3 > Configure > Debug
"""
import logging
import os
from logging import Formatter
from logging.handlers import RotatingFileHandler


FMT = "%(asctime)s %(message)s"


def init(logger_name: str, config: dict, config_dir: str = None):
    level = config["level"].upper()

    logger = logging.getLogger(logger_name)
    logger.propagate = config["propagate"]
    logger.setLevel(level)

    filename = config.get("stderr-file")
    if filename:
        if config_dir:
            filename = os.path.join(config_dir, filename)

        handler = RotatingFileHandler(
            filename,
            config["mode"],
            config["max_bytes"],
            config["backup_count"],
        )

        fmt = Formatter(config["format"])
        handler.setFormatter(fmt)

        logger.addHandler(handler)
